# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Protocol

from dimos.core import Out
from dimos.msgs.sensor_msgs import CameraInfo, Image as ImageMsg, PointCloud2


class Image(Protocol):
    color_image: Out[ImageMsg]


class Camera(Image):
    camera_info: Out[CameraInfo]
    _camera_info: CameraInfo


class DepthCamera(Camera):
    depth_image: Out[ImageMsg]
    depth_camera_info: Out[CameraInfo]


class Pointcloud(Protocol):
    pointcloud: Out[PointCloud2]
