# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Colors:
    GREEN_PRINT_COLOR: str = "\033[32m"
    YELLOW_PRINT_COLOR: str = "\033[33m"
    RED_PRINT_COLOR: str = "\033[31m"
    BLUE_PRINT_COLOR: str = "\033[34m"
    MAGENTA_PRINT_COLOR: str = "\033[35m"
    CYAN_PRINT_COLOR: str = "\033[36m"
    WHITE_PRINT_COLOR: str = "\033[37m"
    RESET_COLOR: str = "\033[0m"
