#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Protocol

from open3d.geometry import PointCloud  # type: ignore[import-untyped]


class PointCloudAccumulator(Protocol):
    def get_point_cloud(self) -> PointCloud:
        """Get the accumulated pointcloud."""
        ...

    def add(self, point_cloud: PointCloud) -> None:
        """Add a pointcloud to the accumulator."""
        ...
