from dimos.perception.detection.detectors import *
from dimos.perception.detection.module2D import (
    Detection2DModule,
)
from dimos.perception.detection.module3D import (
    Detection3DModule,
)
