#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

from dimos.msgs.sensor_msgs import PointCloud2
from dimos.robot.unitree_webrtc.type.lidar import pointcloud2_from_webrtc_lidar
from dimos.utils.testing import SensorReplay


def test_init() -> None:
    lidar = SensorReplay("office_lidar")

    for raw_frame in itertools.islice(lidar.iterate(), 5):
        assert isinstance(raw_frame, dict)
        frame = pointcloud2_from_webrtc_lidar(raw_frame)
        assert isinstance(frame, PointCloud2)
